'''
WDA Element Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBElementCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#f68f7fd9-3a08-4a0b-9253-f1bedf0ae926
'''
import os
import pytest
import unittest
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestElement(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)
        self.temp_file_pic = os.path.join(curPath, 'temp_file_pic.png')

    def tearDown(self):
        self.wda_client.close()
        [os.remove(temp_file) for temp_file in [self.temp_file_pic] if os.path.exists(temp_file)]


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/window/size
     Description: Fetch device window size.
    '''
    def test_window_size(self):
         assert hasattr(self.app.window_size(), 'width') and type(self.app.window_size().width) == int
         assert hasattr(self.app.window_size(), 'height') and type(self.app.window_size().height) == int


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/enabled
     Description: Check element is enbaled or not.
    '''
    def test_btn_enable(self):
        assert self.app(text='ENABLED_BTN').enabled == True

    def test_btn_enable(self):
        assert self.app(text='DISABLED_BTN').enabled == False


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/rect
    '''
    def test_rect(self):
        rect = self.app(text='ENABLED_BTN').bounds
        assert all([
            hasattr(rect, 'x') and type(rect.x) == int,
            hasattr(rect, 'y') and type(rect.y) == int,
            hasattr(rect, 'width') and type(rect.width) == int,
            hasattr(rect, 'height') and type(rect.height) == int,
        ])


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/attribute/{{attribute_name}}
    '''
    def test_attribute_label(self):
        self.assertEqual(self.app(text='ENABLED_BTN').label, 'ENABLED_BTN')

    def test_attribute_invild(self):
        with pytest.raises(AttributeError, match="'Element' object has no attribute 'attribute'"):
            self.app(text='ENABLED_BTN').attribute('invalid_attribute_name')


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/text
    '''
    def test_attribute_btn_text(self):
        self.assertEqual(self.app(text='ENABLED_BTN').text, 'ENABLED_BTN')

    def test_attribute_image_text(self):
        self.assertEqual(self.app(id='IMG_BTN').text, 'applogo')


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/displayed
    '''
    def test_attribute_btn_displayed(self):
        self.assertEqual(self.app(text='ENABLED_BTN').displayed, True)

    def test_attribute_btn_not_displayed(self):
        self.assertEqual(self.app(text='HIDDEN_BTN').displayed, False)


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/selected
    '''
    def test_btn_selected(self):
        self.assertEqual(self.app(text='CHECKED_BTN').selected(), True)

    def test_attribute_btn_not_displayed(self):
        self.assertEqual(self.app(text='UNCHECKED_BTN').selected(), False)


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/name
     NOTE: Return element type.
    '''
    def test_btn_name(self):
        self.assertEqual(self.app(text='CHECKED_BTN').name, 'XCUIElementTypeButton')

    def test_img_name(self):
        self.assertEqual(self.app(text='IMG_BTN').name, 'XCUIElementTypeImage')


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/value
    '''
    def test_input_text(self):
        self.app(id='INPUT_FIELD').set_text('test')
        self.assertEqual(self.app(id='INPUT_FIELD').value, 'test')


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/click
    '''
    def test_input_click(self):
        input_text_val = 'test'
        self.app(id='INPUT_FIELD').set_text(input_text_val)
        assert self.app(id='INPUT_FIELD').value == input_text_val
        self.app(id='CLEAR_INPUT_BTN').click()
        assert self.app(id='INPUT_FIELD').value != input_text_val


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/clear
    '''
    def test_input_clear(self):
        input_text_val = 'test'
        self.app(id='INPUT_FIELD').set_text(input_text_val)
        assert self.app(id='INPUT_FIELD').value == input_text_val
        self.app(id='INPUT_FIELD').clear_text()
        assert self.app(id='INPUT_FIELD').value != input_text_val


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/screenshot
    '''
    def test_screenshot(self):
        self.app.screenshot(png_filename=self.temp_file_pic)
        assert os.path.exists(self.temp_file_pic)


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/screenshot/{{uuid}}
     Description: screenshot for target element.
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/screenshot/{{uuid}}')
    def test_screenshot_element(self):
        pass


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/accessible
    '''
    def test_ele_accessible_is_false(self):
        self.assertFalse(self.app(text='HIDDEN_BTN').accessible)

    def test_ele_accessible_is_true(self):
        self.assertTrue(self.app(text='ENABLED_BTN').accessible)


    '''
    Method: POST 
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/accessibilityContainer
    NOTE Swift accessibility Container is not useful, always return false.
    
    Example:
    ```
    import SwiftUI

    struct ContentView: View {
        var body: some View {
            HStack {
                Text("First Item")
                Divider()
                Text("Second Item")
            }
            .accessibilityElement(children: .combine)
            .accessibility(label: Text("Combined Items"))
        }
    }

    struct ContentView_Previews: PreviewProvider {
        static var previews: some View {
            ContentView()
        }
    }
    ```
    '''
    @pytest.mark.skip('WDA API NOT USEFUL: Not use in SwiftUI.')
    def test_ele_accessibilityContainer_is_true(self):
        self.assertFalse(self.app(id='Combined Items').accessibility_container)

    def test_ele_accessibilityContainer_is_false(self):
        self.assertFalse(self.app(id='ENABLED_BTN').accessibility_container)


    '''
    Method: POST 
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/swipe
    '''
    @pytest.mark.skip('NOT IMPLEMENTED')
    def test_ele_swipe_top(self):
        self.app(text="Go to List").click()
        self.app(text="Row1").click()


    '''
    Method: POST 
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/pinch
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/pinch')
    def test_pinch(self):
        ...


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/touchAndHold
    '''
    def test_touch_and_hold(self):
        try:
            self.app(text='OK').get(timeout=1)
        except wda.exceptions.WDAElementNotFoundError:
            pass
        self.app(text='LONG_TAP_ALERT').tap_hold(duration=2)
        self.assertTrue(self.app(text='LONG_TAP_ALERT_OK').get(timeout=1).displayed) 


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/dragfromtoforduration
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/dragfromtoforduration')
    def test_drag_from_to_for_duration(self):
      pass


    '''
    Method: POST 
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/dragfromtoforduration
    '''
    def test_app_drag_from_to_for_duration(self):
        self.app(text='ListView').click()
        self.assertTrue(self.app(text='Row1').get(timeout=1).displayed) 
        try:
            self.app(text='Row30').get(timeout=1)
        except wda.exceptions.WDAElementNotFoundError:
            pass
        self.app.swipe(500, 800, 500, 200, duration=0.5)
        self.assertTrue(self.app(text='Row30').get(timeout=1).displayed) 


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/pickerwheel/{{uuid}}/select
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/wda/pickerwheel/{{uuid}}/select')
    def test_ele_select(self):
        pass


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/keys
    '''
    def test_wda_keys(self):
        self.app(text='INPUT_FIELD').click()
        self.app.send_keys('hello world')
        self.assertEqual(self.app(text='INPUT_FIELD').value, 'hello world')


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/forceTouch
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/forceTouch')
    def test_ele_force_touch(self):
        pass


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/doubleTap
    '''
    def test_ele_touble_tap(self):
        bounds = self.app(text='DOUBLE_TAP_ALERT').bounds
        x = int(bounds.x + bounds.width * 0.5)
        y = int(bounds.y + bounds.height * 0.5)
        self.app.double_tap(x, y)
        self.assertTrue(self.app(text='DOUBLE_TAP_ALERT_OK').get(timeout=1).displayed)


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/touchAndHold
    '''
    def test_wda_touch_and_hold(self):
        bounds = self.app(text='LONG_TAP_ALERT').bounds
        x = int(bounds.x + bounds.width * 0.5)
        y = int(bounds.y + bounds.height * 0.5)
        self.app.tap_hold(x, y, duration=2)
        self.assertTrue(self.app(text='LONG_TAP_ALERT_OK').get(timeout=1).displayed)



'''
WDA Screenshot Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBScreenshotCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#bf7cb0a1-cc3b-4eb5-a9ee-1e7e63b55df1
'''
import os
import unittest
from typing import List
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestScreenshot(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)
        self.temp_file_pic = os.path.join(curPath, 'temp_file_pic.png')

    def tearDown(self):
        self.app.orientation = wda.PORTRAIT
        self.wda_client.close()
        [os.remove(temp_file) for temp_file in [self.temp_file_pic] if os.path.exists(temp_file)]


    '''
     Method: POST 
     Endpoint: {{baseURL}}/screenshot
    '''
    def test_screenshot(self):
        self.app.screenshot(self.temp_file_pic)
        assert os.path.exists(self.temp_file_pic)



'''
WDA Orientation Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBOrientationCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#1ca27827-9931-4315-a6aa-141b6015ac04
'''
import os
import time
import pytest
import unittest
from typing import List
from collections.abc import Iterable
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestOrientation(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)
        self.temp_file_pic = os.path.join(curPath, 'temp_file_pic.png')

    def tearDown(self):
        self.app.orientation = wda.PORTRAIT
        self.wda_client.close()
        [os.remove(temp_file) for temp_file in [self.temp_file_pic] if os.path.exists(temp_file)]


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/orientation
    '''
    def test_orientation(self):
        assert self.app.orientation in ['PORTRAIT', 'LANDSCAPE']


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/orientation
    '''
    def test_set_orientation(self):
        self.app.orientation = wda.LANDSCAPE
        time.sleep(1)
        assert self.app.orientation == 'LANDSCAPE'
    
    
    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/rotation
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/session/{{sessionId}}/rotation')
    def test_get_rotation(self):
        pass


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/rotation
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [POST] {{baseURL}}/session/{{sessionId}}/rotation')
    def test_set_rotation(self):
        pass




'''
WDA Custom Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBCustomCommands.m#L38
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#573d11c5-c434-4753-b845-4a3afcc4b614
'''
import os
import pytest
import unittest
import jsonschema
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestDevice(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)

    def tearDown(self):
        self.wda_client.close()

    '''
    Method: POST
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/deactivateApp
    Description: Put app into background and than put it back.
    '''
    def test_deactivate_app(self):
        self.app.deactivate(duration=1)


    '''
    Method: POST
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/keyboard/dismiss
    Description: Put the keyboard into the background.
    '''
    def test_keybord_dismiss(self):
        with pytest.raises(RuntimeError) as e:
            self.app.keyboard_dismiss()
    
    
    '''
    Method: POST
    Endpoint: {{baseURL}}/wda/lock
    Endpoint: {{baseURL}}/wda/unlock
    Description: Lock the device.
    '''
    def test_keybord_lock_and_unlock(self):
        self.app.lock()
        self.app.unlock()


    '''
    Method: GET
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/screen
    Description: UIKit scale factor
    Refs:
        https://developer.apple.com/library/archive/documentation/DeviceInformation/Reference/iOSDeviceCompatibility/Displays/Displays.html
    There is another way to get scale
        self._session_http.get("/wda/screen").value returns {"statusBarSize": {'width': 320, 'height': 20}, 'scale': 2}
    '''
    def test_scale(self):
        self.assertIsInstance(self.app.scale, int)
        
    
    '''
    Method: GET
    Endpoint: {{baseURL}}/wda/activeAppInfo
    Description: Return bundleId pid and etc. like:
    {'processArguments': {'env': {}, 'args': []}, 'name': '', 'pid': 19052, 'bundleId': 'com.test.cert.TestCert'}
    '''
    def test_active_app_info(self):
        except_json_schema = {"$schema":"http://json-schema.org/draft-07/schema#","type":"object","properties":
                              {"processArguments":{"type":"object","properties":{"env":{"type":"object",
                                "additionalProperties":False},"args":{"type":"array"}},"additionalProperties":False,
                                "required":["env","args"]},"name":{"type":"string"},"pid":{"type":"integer"},
                                "bundleId":{"type":"string"}},"additionalProperties":False,"required":
                                ["processArguments","name","pid","bundleId"]}
        self.assertTrue(jsonschema.Draft7Validator(except_json_schema).is_valid(self.app.app_current()))


    '''
    Method: POST
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/setPasteboard
    Description: Set paste board  board.
    '''
    def test_set_paste_board(self):
        self.app.set_clipboard('test')

    '''
    Method: POST
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/getPasteboard
    #NOTE: IS NOT USED. RETURN NULL.
    e.g:
    curl http://127.0.0.1:8100/session/3D28C745-5290-4787-948C-43C7A27E6146/wda/getPasteboard -X POST -v
    *   Trying 127.0.0.1:8100...
    * Connected to 127.0.0.1 (127.0.0.1) port 8100 (#0)
    > POST /session/3D28C745-5290-4787-948C-43C7A27E6146/wda/getPasteboard HTTP/1.1
    > Host: 127.0.0.1:8100
    > User-Agent: curl/7.86.0
    > Accept: */*
    > 
    * Mark bundle as not supporting multiuse
    < HTTP/1.1 400 Bad Request
    < Server: WebDriverAgent/1.0
    < Access-Control-Allow-Origin: *
    < Date: Thu, 21 Mar 2024 02:18:52 GMT
    < Access-Control-Allow-Headers: Content-Type, X-Requested-With
    < Accept-Ranges: bytes
    < Content-Length: 0
    < Connection: close
    < 
    * Closing connection 0
    '''
    @pytest.mark.skip('WDA API NOT USEFUL: {{baseURL}}/session/{{sessionId}}/wda/getPasteboard')
    def test_get_paste_board(self):
        '''Wait to PR merge: https://github.com/openatx/facebook-wda/pull/133/files'''
        ...


    '''
    Method: GET
    Endpoint: {{baseURL}}/wda/device/info
    Description: Return device info.
    Example Return:
    {"timeZone": "GMT+0800", "currentLocale": "zh_CN", "model": "iPhone", "uuid": 
    "25E3142B-303E-41FE-9F6A-2C303CB66FBC", "thermalState": 1, "userInterfaceIdiom": 0,
    "userInterfaceStyle": "light", "name": "iPhone", "isSimulator": False}
    '''
    def test_device_info(self):
        expect_schema = {"$schema":"http://json-schema.org/draft-07/schema#","type":"object",
                         "properties":{"timeZone":{"type":"string"},"currentLocale":{"type":"string"},
                        "model":{"type":"string"},"uuid":{"type":"string"},"thermalState":{"type":"integer"},
                        "userInterfaceIdiom":{"type":"integer"},"userInterfaceStyle":{"type":"string"},
                        "name":{"type":"string"},"isSimulator":{"type":"boolean"}},"additionalProperties":False,
                        "required":["timeZone","currentLocale","model","uuid","thermalState","userInterfaceIdiom",
                        "userInterfaceStyle","name","isSimulator"]}
        self.assertTrue(jsonschema.Draft7Validator(expect_schema).is_valid(self.app.device_info()))


    '''
    Method: GET
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/batteryInfo
    Description: Return device battery info.
    Example Return:
    {"level": 0.5799999833106995, "state": 2}
    '''
    def test_battery_info(self):
        import json
        except_json_schema = {"$schema":"http://json-schema.org/draft-07/schema#","type":"object",
                              "properties":{"level":{"type":"number"},"state":{"type":"integer"}},
                              "additionalProperties":False,"required":["level","state"]}
        self.assertTrue(jsonschema.Draft7Validator(except_json_schema).is_valid(self.app.battery_info()))


    '''
    Method: GET
    Endpoint: {{baseURL}}/wda/homescreen
    Description: back to home screen
    '''
    def test_homescreen(self):
        self.assertEqual(self.app.app_current().get('bundleId'), UNDER_TEST_BUNDLE_ID)
        self.app.home()
        self.assertEqual(self.app.app_current().get('bundleId'), HOME_SCREEN_BUNDLE_ID)


    '''
    Method: GET
    Endpoint: {{baseURL}}/wda/locked
    Description: check device is locked or not.
    '''
    def test_locked(self):
        self.assertFalse(self.app.locked())



'''
WDA ALERT Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBAlertViewCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#c6d397f5-8ece-4f2c-8bfd-26e798f83cf2
'''
import os
import pytest
import unittest
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestAlert(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)

    def tearDown(self):
        self.wda_client.close()


    '''
    Method: GET 
    Endpoint: {{baseURL}}/alert/text
    Description: Get the content of the Alert (only the content, not including options), 
    an exception(wda.exceptions.WDARequestError) will be thrown if no Alert is present.
    '''

    def test_alert_text_endpoint_confirmation(self):
        self.app(label='ACCEPT_OR_REJECT_ALERT').click()
        self.assertEqual('Confirmation\nDo you accept?', self.wda_client.alert.text)
    
    def test_alert_text_endpoint_when_no_alert(self):
        with pytest.raises(wda.exceptions.WDARequestError,  match="status=110, value={'error': \'no such alert', "\
                           "'message': 'An attempt was made to operate on a modal dialog when one was not open'}"):
            self.wda_client.alert.text


    '''
    Method: POST
    Endpoint: {{baseURL}}/session/{{sessionId}}/alert/text
    '''
    def test_alert_text_input(self):
        with pytest.raises(wda.exceptions.WDARequestError,  match="status=110, value={'error': \'no such alert', "\
                           "'message': 'An attempt was made to operate on a modal dialog when one was not open'}"):
            self.wda_client.alert.set_text('hello world')


    '''1
    Method: GET 
    Endpoint: {{baseURL}}/session/{{sessionId}}/wda/alert/buttons
    Description: Get buttons for all prompt alert buttons.
    '''
    def test_alert_text_button_endpoint_value(self):
        self.app(label=ALERT_ENUMS.CONFIRM.value).click()
        self.assertEqual(['Reject', 'Accept'], self.wda_client.alert.buttons())


    '''
    Method: POST 
    Endpoint: {{baseURL}}/alert/dismiss
    Description: Dimiss the alert which display.
    '''
    def test_alert_text_dimiss_endpoint(self):
        self.app(label=ALERT_ENUMS.CONFIRM.value).click()
        self.app(text='Reject').get(timeout=1)
        self.wda_client.alert.dismiss()
        try:
            self.app(text='Reject').get(timeout=1)
        except wda.exceptions.WDAElementNotFoundError:
            return
        raise AssertionError('Alert not dismissed')


    '''
    Method: POST 
    Endpoint: {{baseURL}}/alert/accept
    Description: Accept the alert which display.
    '''
    def test_alert_text_accept_endpoint(self):
        self.app(label=ALERT_ENUMS.CONFIRM.value).click()
        self.app.alert.accept()
        try:
            self.app(text='Accept').get(timeout=1)
        except wda.exceptions.WDAElementNotFoundError:
            return
        raise AssertionError('Alert not dismissed')



# coding: utf-8

import functools
import typing
import inspect


def inject_call(fn, *args, **kwargs):
    """
    Call function without known all the arguments

    Args:
        fn: function
        args: arguments
        kwargs: key-values
    
    Returns:
        as the fn returns
    """
    assert callable(fn), "first argument must be callable"

    st = inspect.signature(fn)
    fn_kwargs = {
        key: kwargs[key]
        for key in st.parameters.keys() if key in kwargs
    }
    ba = st.bind(*args, **fn_kwargs)
    ba.apply_defaults()
    return fn(*ba.args, **ba.kwargs)


def limit_call_depth(n: int):
    """
    n = 0 means not allowed recursive call
    """
    def wrapper(fn: typing.Callable):
        if not hasattr(fn, '_depth'):
            fn._depth = 0

        @functools.wraps(fn)
        def _inner(*args, **kwargs):
            if fn._depth > n:
                raise RuntimeError("call depth exceed %d" % n)

            fn._depth += 1
            try:
                return fn(*args, **kwargs)
            finally:
                fn._depth -= 1
        
        _inner._fn = fn
        return _inner
    
    return wrapper


def main():
    @limit_call_depth(1)
    def foo(a):
        print("Foo:", a)
        if a == 2:
            return
        foo(a+1)

    foo(0)

if __name__ == "__main__":
    main()


# Created: codeskyblue 2020/05
# Inspired by: https://github.com/iOSForensics/pymobiledevice

import logging
import os
import plistlib
import socket
import struct
import sys
from functools import lru_cache
from typing import Optional, Union

from .exceptions import MuxConnectError, MuxError


PROGRAM_NAME = "facebook-wda"
logger = logging.getLogger(PROGRAM_NAME)


class SafeStreamSocket():
    def __init__(self, addr: Union[str, tuple, socket.socket]):
        """
        Args:
            addr: can be /var/run/usbmuxd or (localhost, 27015)
        """
        self._sock = None
        if isinstance(addr, socket.socket):
            self._sock = addr
            return
        if isinstance(addr, str):
            if not os.path.exists(addr):
                raise MuxError("socket unix:{} unable to connect".format(addr))
            family = socket.AF_UNIX
        else:
            family = socket.AF_INET

        self._sock = socket.socket(family, socket.SOCK_STREAM)
        self._sock.connect(addr)

    def recvall(self, size: int) -> bytearray:
        buf = bytearray()
        while len(buf) < size:
            chunk = self._sock.recv(size - len(buf))
            if not chunk:
                raise MuxError("socket connection broken")
            buf.extend(chunk)
        return buf

    def sendall(self, data: Union[bytes, bytearray]) -> int:
        return self._sock.sendall(data)

    def close(self):
        logger.debug("Socket closed")
        self._sock.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class PlistSocket(SafeStreamSocket):
    def __init__(self, addr: str, tag: int = 0):
        super().__init__(addr)
        self._tag = tag
        self._first = True

    def send_packet(self, payload: dict, reqtype: int = 8):
        """
        Args:
            payload: required

            # The following args only used in the first request
            reqtype: request type, always 8 
            tag: int
        """
        body_data = plistlib.dumps(payload)
        if self._first:  # first package
            length = 16 + len(body_data)
            header = struct.pack(
                "IIII", length, 1, reqtype,
                self._tag)  # version: 1, request: 8(?), tag: 1(?)
        else:
            header = struct.pack(">I", len(body_data))
        self.sendall(header + body_data)

    def recv_packet(self, header_size=None) -> dict:
        if self._first or header_size == 16:  # first receive
            header = self.recvall(16)
            (length, version, resp, tag) = struct.unpack("IIII", header)
            length -= 16  # minus header length
            self._first = False
        else:
            header = self.recvall(4)
            (length, ) = struct.unpack(">I", header)

        body_data = self.recvall(length)
        payload = plistlib.loads(body_data)
        return payload

    def send_recv_packet(self, payload: dict) -> dict:
        self.send_packet(payload)
        return self.recv_packet()


class Usbmux:
    def __init__(self, address: Optional[Union[str, tuple]] = None):
        if address is None:
            if os.name == "posix":  # linux or darwin
                address = "/var/run/usbmuxd"
            elif os.name == "nt":  # windows
                address = ('127.0.0.1', 27015)
            else:
                raise EnvironmentError("Unsupported system:", sys.platform)

        self.__address = address
        self.__tag = 0

    def _next_tag(self) -> int:
        self.__tag += 1
        return self.__tag

    def create_connection(self) -> PlistSocket:
        return PlistSocket(self.__address, self._next_tag())

    def send_recv(self, payload: dict) -> dict:
        with self.create_connection() as s:
            s.send_packet(payload)
            recv_data = s.recv_packet()
            self._check(recv_data)
            return recv_data

    def device_list(self):
        """
        Return example:
        [{'ConnectionSpeed': 480000000,
          'ConnectionType': 'USB',
          'DeviceID': 37,
          'LocationID': 123456,
          'ProductID': 4776,
          'SerialNumber': 'xxx',
          'UDID': 'xxx',
          'USBSerialNumber': 'xxx'}}]}
        """
        payload = {
            "MessageType": "ListDevices",  # 必选
            "ClientVersionString": "libusbmuxd 1.1.0",
            "ProgName": PROGRAM_NAME,
            "kLibUSBMuxVersion": 3,
            # "ProcessID": 0, # Xcode send it processID
        }
        data = self.send_recv(payload)
        _devices = [item['Properties'] for item in data['DeviceList']]
        return [d for d in _devices if d['ConnectionType'] == 'USB']

    def get_single_device_udid(self) -> str:
        """ to run this function, there must be only one device connected
        
        Raises:
            MuxError
        """
        devices = self.device_list()
        if len(devices) == 0:
            raise MuxError("No device connected")
        if len(devices) > 1:
            raise MuxError("More then two device connected")
        udid = devices[0]['SerialNumber']
        return udid

    @lru_cache(1024)
    def device(self, udid: str) -> "Device":
        """ return device object """
        return Device(udid, self)

    def _check(self, data: dict):
        if 'Number' in data and data['Number'] != 0:
            raise MuxError(data)

    def read_system_BUID(self):
        """ BUID is always same """
        data = self.send_recv({
            'ClientVersionString': 'libusbmuxd 1.1.0',
            'MessageType': 'ReadBUID',
            'ProgName': PROGRAM_NAME,
            # 'kLibUSBMuxVersion': 3
        })
        return data['BUID']


class Device(object):
    def __init__(self, udid: str, _usbmux=None):
        assert udid, "udid should not empty"
        self._usbmux = _usbmux or Usbmux()
        self._udid = udid
        self._info = self.info

    # DeviceID will be changed if device re-plug
    # So can not use cached_property here
    @property
    def info(self) -> dict:
        """
        Example return:
        {
            "SerialNumber": "xxxx", # udid
            "DeviceID": 12,
        }
        """
        for dinfo in self._usbmux.device_list():
            if dinfo['SerialNumber'] == self._udid:
                return dinfo

        raise MuxError("Device {} not connected".format(self._udid))

    def create_inner_connection(self, port: int = 0xf27e) -> PlistSocket:
        # I really donot know why do this
        # The following code convert port(0x1234) to port(0x3412)
        _port = ((port & 0xff) << 8) | (port >> 8)
        logger.debug("port convert %s(%d) -> %s(%d)", hex(port), port,
                     hex(_port), _port)
        _original_port = port
        del (port)

        device_id = self.info['DeviceID']
        conn = self._usbmux.create_connection()
        payload = {
            'DeviceID': device_id,  # Required
            'MessageType': 'Connect',  # Required
            'PortNumber': _port,  # Required
            'ProgName': PROGRAM_NAME,
        }
        logger.debug("Send payload: %s", payload)
        data = conn.send_recv_packet(payload)
        logger.debug("connect port: %d", _port)
        if 'Number' in data and data['Number'] != 0:
            err_code = data['Number']
            if err_code == 3:
                raise MuxConnectError(
                    "device port:{} is not ready".format(_original_port))
            else:
                raise MuxError(data)
        logger.debug("connection established")
        return conn



'''
WDA Session Command Testing
API FBSessionCommands.m
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBSessionCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#f94ebe16-ae2a-4098-879e-570a0138c7f4
'''
import os
import pytest
import unittest
import jsonschema
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestSessionCommands(unittest.TestCase):

    def setUp(self):
        self.wda_client: wda.Client = wda.Client()
        # self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)
        self.temp_file_pic = os.path.join(curPath, 'temp_file_pic.png')

    def tearDown(self):
        self.wda_client.close()
        [os.remove(temp_file) for temp_file in [self.temp_file_pic] if os.path.exists(temp_file)]


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/url
    '''
    @pytest.mark.skip("UNKNOW HOW TO TEST: [POST] {{baseURL}}/session/{{sessionId}}/url")
    def test_url(self):
        pass


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session
    '''
    def test_create_session_id(self):
        before_create_session_id = self.wda_client.session_id
        self.assertIsNot(self.wda_client.session().session_id, before_create_session_id, None)
        client: wda.Client = wda.Client().session(bundle_id=UNDER_TEST_BUNDLE_ID)


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/apps/launch
    '''
    def test_launch_app(self):
        self.wda_client.app_activate(UNDER_TEST_BUNDLE_ID)
        self.wda_client.app_current().get('bundle_id') == UNDER_TEST_BUNDLE_ID


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/apps/list
    '''
    def test_app_list(self):
        self.wda_client.session(UNDER_TEST_BUNDLE_ID)
        self.assertIsInstance(self.wda_client.app_list(), list)
        self.assertEqual(self.wda_client.app_list()[0].get('bundleId'), UNDER_TEST_BUNDLE_ID)

    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/apps/state
    '''
    def test_app_state(self):
        self.wda_client.session(UNDER_TEST_BUNDLE_ID)
        self.assertIsInstance(self.wda_client.app_state(UNDER_TEST_BUNDLE_ID), dict)
        self.assertEqual(self.wda_client.app_state(UNDER_TEST_BUNDLE_ID).get('value'), 4)

    def test_unknow_app_state(self):
        self.assertEqual(self.wda_client.app_state(UNKNOW_BUNDLE_ID).get('value'), 1)


    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/apps/terminate
    '''
    def test_app_terminate(self):
        self.wda_client.app_terminate(UNDER_TEST_BUNDLE_ID)
        self.assertTrue(self.wda_client.app_state(UNDER_TEST_BUNDLE_ID).get('value') == 1)


    '''
     Method: GET 
     Endpoint: {{baseURL}}/status
    '''
    def test_status_command_return_schema(self):
        '''
        Current Origin Status Return:
        {
            "build": {
                "time": "Mar 18 2024 11:29:21",
                "productBundleIdentifier": "com.facebook.WebDriverAgentRunner"
            },
            "os": {
                "testmanagerdVersion": 28,
                "name": "iOS",
                "sdkVersion": "16.4",
                "version": "16.3.1"
            },
            "device": "iphone",
            "ios": {
                "ip": "XX.XX.XX.XX"
            },
            "message": "WebDriverAgent is ready to accept commands",
            "state": "success",
            "ready": true,
            "sessionId": "XXXXX"
        }
        '''
        self.wda_client.session(UNDER_TEST_BUNDLE_ID)
        expect_schema = {"$schema":"http://json-schema.org/draft-07/schema#","type":"object","properties":{"build":\
                        {"type":"object","properties":{"time":{"type":"string"},"productBundleIdentifier":\
                        {"type":"string"}},"additionalProperties":"false","required":["time",\
                        "productBundleIdentifier"]},"os":{"type":"object","properties":{"testmanagerdVersion"\
                        :{"type":"integer"},"name":{"type":"string"},"sdkVersion":{"type":"string"},"version":\
                        {"type":"string"}},"additionalProperties":"false","required":["testmanagerdVersion","name",\
                        "sdkVersion","version"]},"device":{"type":"string"},"ios":{"type":"object","properties":\
                        {"ip":{"type":"string"}},"additionalProperties":"false","required":["ip"]},"message":\
                        {"type":"string"},"state":{"type":"string"},"ready":{"type":"boolean"},"sessionId":\
                        {"type":"string"}},"additionalProperties":"false","required":["build","os","device","ios",\
                        "message","state","ready", "sessionId"]}

        self.assertTrue(jsonschema.Draft7Validator(expect_schema).is_valid(self.wda_client.status()))

    def test_status_command_state_and_ready(self):
        status = self.wda_client.status()
        assert all([status['state'] == 'success', status['ready'] is True])


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/session/{{sessionId}}')
    def test_session_info(self):
        pass

    '''
     Method: DELETE 
     Endpoint: {{baseURL}}/session/{{sessionId}}
    '''
    def test_delete_session(self):
        self.wda_client.session(UNDER_TEST_BUNDLE_ID)
        assert self.wda_client.session_id is not None
        print(self.wda_client.close())

    '''
     Method: POST 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/apps/activate
    '''
    def test_app_activate(self):
        self.wda_client.session()
        self.wda_client.app_activate(UNDER_TEST_BUNDLE_ID)
        self.assertEqual(self.wda_client.app_current().get('bundleId'), UNDER_TEST_BUNDLE_ID)


    '''
     Method: GET 
     Endpoint: {{baseURL}}/wda/healthcheck
    '''
    def test_health_check(self):
        health_check = self.wda_client.healthcheck()
        assert all([
            hasattr(health_check, 'value'),
            hasattr(health_check, 'sessionId'),
            health_check.get('value') == None,
            health_check.get('sessionId') == None
        ])
    
    def test_health_check_has_session_id(self):
        self.wda_client.session(UNDER_TEST_BUNDLE_ID)
        health_check = self.wda_client.healthcheck()
        assert all([
            hasattr(health_check, 'value'),
            hasattr(health_check, 'sessionId'),
            health_check.get('value') == None,
            health_check.get('sessionId') == self.wda_client.session_id
        ])

    '''
     Method: GET
     Endpoint: {{baseURL}}/session/{{sessionId}}/appium/settings
     Return Example:

     ```
     {
        "mjpegFixOrientation": false,
        "boundElementsByIndex": false,
        "mjpegServerFramerate": 10,
        "screenshotOrientation": "auto",
        "reduceMotion": false,
        "elementResponseAttributes": "type,label",
        "screenshotQuality": 3,
        "mjpegScalingFactor": 100,
        "keyboardPrediction": 0,
        "defaultActiveApplication": "auto",
        "mjpegServerScreenshotQuality": 25,
        "defaultAlertAction": "",
        "keyboardAutocorrection": 0,
        "useFirstMatch": false,
        "shouldUseCompactResponses": true,
        "customSnapshotTimeout": 15,
        "dismissAlertButtonSelector": "",
        "activeAppDetectionPoint": "64.00,64.00",
        "snapshotMaxDepth": 50,
        "waitForIdleTimeout": 10,
        "includeNonModalElements": false,
        "acceptAlertButtonSelector": "",
        "animationCoolOffTimeout": 2
    }
    ```
    '''
    def test_appium_settings(self):
        appium_settings = self.wda_client.appium_settings()
        expect_schema = {"$schema":"http://json-schema.org/draft-07/schema#","type":"object","properties":
                         {"mjpegFixOrientation":{"type":"boolean"},"boundElementsByIndex":{"type":"boolean"},
                          "mjpegServerFramerate":{"type":"integer"},"screenshotOrientation":{"type":"string"},
                          "reduceMotion":{"type":"boolean"},"elementResponseAttributes":{"type":"string"},
                          "screenshotQuality":{"type":"integer"},"mjpegScalingFactor":{"type":"integer"},
                          "keyboardPrediction":{"type":"integer"},"defaultActiveApplication":{"type":"string"},
                          "mjpegServerScreenshotQuality":{"type":"integer"},"defaultAlertAction":{"type":"string"},
                          "keyboardAutocorrection":{"type":"integer"},"useFirstMatch":{"type":"boolean"},
                          "shouldUseCompactResponses":{"type":"boolean"},"customSnapshotTimeout":{"type":"integer"},
                          "dismissAlertButtonSelector":{"type":"string"},"activeAppDetectionPoint":{"type":"string"},
                          "snapshotMaxDepth":{"type":"integer"},"waitForIdleTimeout":{"type":"integer"},
                          "includeNonModalElements":{"type":"boolean"},"acceptAlertButtonSelector":{"type":"string"},
                          "animationCoolOffTimeout":{"type":"integer"}},"additionalProperties":False,
                          "required":["mjpegFixOrientation","boundElementsByIndex","mjpegServerFramerate",
                            "screenshotOrientation","reduceMotion","elementResponseAttributes","screenshotQuality",
                            "mjpegScalingFactor","keyboardPrediction","defaultActiveApplication","mjpegServerScreenshotQuality",
                            "defaultAlertAction","keyboardAutocorrection","useFirstMatch","shouldUseCompactResponses",
                            "customSnapshotTimeout","dismissAlertButtonSelector","activeAppDetectionPoint","snapshotMaxDepth",
                            "waitForIdleTimeout","includeNonModalElements","acceptAlertButtonSelector","animationCoolOffTimeout"]} 
        self.assertTrue(jsonschema.Draft7Validator(expect_schema).is_valid(appium_settings))   


    '''
     Method: GET
     Endpoint: {{baseURL}}/wda/shutdown
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/wda/shutdown')
    def test_wda_shutdown(self):
        pass


    '''
     Method: GET
     Endpoint: {{baseURL}}/health
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/health')
    def test_health(self):
        pass



'''
WDA Find Element Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBElementCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#56e19c88-8571-48d3-a9f1-8e9bd0cbae0d
'''
import os
import pytest
import unittest
from typing import List
from collections.abc import Iterable
import wda
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))


class TestFindElement(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)
        self.temp_file_pic = os.path.join(curPath, 'temp_file_pic.png')

    def tearDown(self):
        self.wda_client.close()
        [os.remove(temp_file) for temp_file in [self.temp_file_pic] if os.path.exists(temp_file)]


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/elements
    '''
    def test_elements(self):
        elements = self.app(text='IMG_BTN').find_elements()
        self.assertTrue(isinstance(elements, Iterable))  # 检查是否可迭代
        for element in elements:
            self.assertIsInstance(element, wda.Element) 

    def test_element_ids(self):
        elements = self.app(text='IMG_BTN').find_element_ids()
        self.assertTrue(isinstance(elements, Iterable))  # 检查是否可迭代
        for element in elements:
            self.assertIsInstance(element, str) 


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/{{uuid}}/elements
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] /session/{{sessionId}}/element/{{uuid}}/elements')
    def test_ele_elements(self):
        pass


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/session/{{sessionId}}/element')
    def test_wda_element(self):
        pass


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/getVisibleCells
    
    Return Example:
    ```
        {'value': [{'ELEMENT': '31000000-0000-0000-996B-000000000000', 
        'element-6066-11e4-a52e-4f735466cecf': '31000000-0000-0000-996B-000000000000'}], 
        'sessionId': 'BC8DE836-6F48-4F7B-B540-FCEC97C06068', 'status': 0}
    ```
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/session/{{sessionId}}/wda/element/{{uuid}}/getVisibleCells')
    def test_wda_element(self):
        self.app(text='ListView').click()
        ele: wda.Element = self.app(label='LIST_CONTAINER').get(timeout=1)
        visible_cells_response = ele.http.get(f'/wda/element/{ele.id}/getVisibleCells')


    '''
     Method: GET 
     Endpoint: {{baseURL}}/session/{{sessionId}}/element/active
    '''
    @pytest.mark.skip('NOT IMPLEMENTED: [GET] {{baseURL}}/session/{{sessionId}}/element/active')
    def test_wda_active(self):
        pass



'''
WDA DEBUG Command Testing
source code here: https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBDebugCommands.m
WDA API document and example(not offical): https://documenter.getpostman.com/view/1837823/TVmMhJNB#a17feaf2-237e-4fe7-8e2f-75f132420d26
'''
import os
import pytest
import unittest
import jsonschema
import wda
from lxml import etree
from .constant import *

curPath = os.path.abspath(os.path.dirname(__file__))

#TODO : 更新断言在apprelease后

class TestDebug(unittest.TestCase):

    def setUp(self):
        self.under_test_bundle_id = UNDER_TEST_BUNDLE_ID
        self.wda_client: wda.Client = wda.Client()
        self.app = self.wda_client.session(bundle_id=self.under_test_bundle_id)

    def tearDown(self):
        self.wda_client.close()

    
    '''
    Method: GET 
    Endpoint: {{baseURL}}/source
    Description: Fetch the source tree (DOM) of the current page.
    '''
    def test_source(self):
        self.app(text='ListView').click()
        source_tree_xml_str: str = self.app.source()
        xml_bytes = source_tree_xml_str.encode('utf-8')
        self.assertTrue(etree.fromstring(xml_bytes))


    '''
    Method: GET 
    Endpoint: {{baseURL}}/wda/accessibleSource
    '''
    def test_accessible_source(self):
        self.app(text='ListView').click()
        source = self.app.source(accessible=True)
        self.assertIsInstance(source, dict)
        assert 'name' in source, "'name' field is missing"
        assert 'type' in source, "'type' field is missing"
        assert 'children' in source, "'children' field is missing"


from enum import Enum


UNDER_TEST_BUNDLE_ID = "com.test.cert.TestCert"

UNKNOW_BUNDLE_ID = "com.apple.unknown"

HOME_SCREEN_BUNDLE_ID = "com.apple.springboard"

class ALERT_ENUMS(Enum):
    CONFIRM = "ACCEPT_OR_REJECT_ALERT"
    INPUT = "INPUT_ALERT"
    

